//
// ********************************************************************
// * License and Disclaimer                                           *
// *                                                                  *
// * The  Geant4 software  is  copyright of the Copyright Holders  of *
// * the Geant4 Collaboration.  It is provided  under  the terms  and *
// * conditions of the Geant4 Software License,  included in the file *
// * LICENSE and available at  http://cern.ch/geant4/license .  These *
// * include a list of copyright holders.                             *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.  Please see the license in the file  LICENSE  and URL above *
// * for the full disclaimer and the limitation of liability.         *
// *                                                                  *
// * This  code  implementation is the result of  the  scientific and *
// * technical work of the GEANT4 collaboration.                      *
// * By using,  copying,  modifying or  distributing the software (or *
// * any work based  on the software)  you  agree  to acknowledge its *
// * use  in  resulting  scientific  publications,  and indicate your *
// * acceptance of all terms of the Geant4 Software license.          *
// ********************************************************************
//
#include "globals.hh"

#include "VKShield.hh"
#include "VKUtil.hh"

#include "G4LogicalSkinSurface.hh"
#include "G4LogicalBorderSurface.hh"
#include "G4SystemOfUnits.hh"
#include "G4Cons.hh"
#include "TMath.h"


VKShield::VKShield(G4RotationMatrix *pRot,
                             const G4ThreeVector &tlate,
                             G4LogicalVolume *pMother_log,
                             G4bool pMany,
                             G4int pCopyNo,
                             const VKDetectorConstruction* pDetConstr)

 // Pass info to the G4PVPlacement constructor
 : G4PVPlacement(pRot, 
                 tlate,
                 CreateTemporaryLogicalVolume(), // <-- see VKUtil.cpp
                 "shield", 
                 pMother_log, 
                 pMany, 
                 pCopyNo),
   fDetConstr(pDetConstr),              
   fMother_log(pMother_log)
{

    // default dimensions    
    SetDefaults();

    // List of materials
    fMaterials = VKMaterials::GetInstance();

    // Get materials
    G4Material* alu   = fMaterials -> GetMaterial("ALUMINIUM");
    G4Material* brass = fMaterials -> GetMaterial("BRASS");
    G4Material* pmma  = fMaterials -> GetMaterial("G4_PLEXIGLASS");
    G4Material* labVacuum = fMaterials -> GetMaterial("LABVACUUM");
    assert(alu != nullptr); // check that material exists
    assert(brass != nullptr);
    assert(pmma != nullptr);
    assert(labVacuum != nullptr);
        
    // rotation matrices, translation and transformation
    G4RotationMatrix doNotRotate; 
    doNotRotate.rotateX(0.);
    G4double displ, startAngle, spanningAngle;
    G4ThreeVector translation; 
    G4Transform3D transform;


    // -------------------------------------------- //
    //             (Imaginary) container            //
    // -------------------------------------------- //
    
    // *** All other volumes should fit inside this volume !

    // solid
    fContainer_sol = new G4Tubs("shield_sol", 0., fContainerDiameter/2, 
                              fContainerLength/2, 0.*deg, 360.*deg);

    // logical
    fContainer_log = new G4LogicalVolume(fContainer_sol, labVacuum, "shield_log");
    
    
    // ------------------------------- //
    //            ALU FRAME            //
    // ------------------------------- //

    G4double aluFrameInnerRadius = fAluFrameOuterDiameter/2 - fAluFrameThickness;
    fAluFrame_sol = new G4Tubs("shieldAluFrame_sol",
                               aluFrameInnerRadius, 
                               fAluFrameOuterDiameter/2,
                               fAluFrameLength/2,
                               startAngle=0*deg, 
                               spanningAngle=360*deg);
  
    fAluFrame_log = new G4LogicalVolume(fAluFrame_sol, alu, "shieldAluFrame_log");
  
    new G4PVPlacement(0, G4ThreeVector(0, 0, 0), fAluFrame_log, 
                      "shieldAluFrame", fContainer_log, false, 0, fCheckOverlaps);


    // ------------------------------- //
    //             WINDOW              //
    // ------------------------------- //

    fWindow_sol = new G4Tubs("shieldWindow_sol", 
                             0., 
                             aluFrameInnerRadius,
                             fWindowLength/2,
                             startAngle=0*deg, 
                             spanningAngle=360*deg);
  
    fWindow_log = new G4LogicalVolume(fWindow_sol, pmma, "shieldWindow_log");
  
    displ = -(fAluFrameLength - fWindowLength)/2;
    new G4PVPlacement(0, G4ThreeVector(0, 0, displ), fWindow_log, 
                      "shieldWindow", fContainer_log, false, 0, fCheckOverlaps);


    // ------------------------------- //
    //             ENDCAP              //
    // ------------------------------- //

    fEndcap_sol = new G4Tubs("shieldEndcap_sol", 
                             0., 
                             aluFrameInnerRadius,
                             fEndcapLength/2,
                             startAngle=0*deg, 
                             spanningAngle=360*deg);
  
    fEndcap_log = new G4LogicalVolume(fEndcap_sol, pmma, "shieldEndcap_log");
  
    displ = (fAluFrameLength - fEndcapLength)/2;
    new G4PVPlacement(0, G4ThreeVector(0, 0, displ), fEndcap_log, 
                      "shieldEndcap", fContainer_log, false, 0, fCheckOverlaps);


    // ------------------------------- //
    //        HEAVY ELEMENT 1          //
    // ------------------------------- //

    fHeavyElem1_sol = new G4Tubs("shieldHeavyElem1_sol", 
                                 0., 
                                 aluFrameInnerRadius,
                                 fHeavyElem1Length/2,
                                 startAngle=0*deg, 
                                 spanningAngle=360*deg);
      
    fHeavyElem1_log = new G4LogicalVolume(fHeavyElem1_sol, brass, "shieldHeavyElem1_log");
  
    displ = (fAluFrameLength - fHeavyElem1Length)/2 - fEndcapLength;
    new G4PVPlacement(0, G4ThreeVector(0, 0, displ), fHeavyElem1_log, 
                      "shieldHeavyElem1", fContainer_log, false, 0, fCheckOverlaps);


    // ------------------------------- //
    //             WING 1              //
    // ------------------------------- //

    fWing1_sol = new G4Tubs("shieldWing1_sol", 
                             fAluFrameOuterDiameter/2, 
                             fWing1OuterDiameter/2,
                             fWing1Length/2,
                             startAngle=0*deg, 
                             spanningAngle=360*deg);
  
    fWing1_log = new G4LogicalVolume(fWing1_sol, pmma, "shieldWing1_log");
  
    displ = fWing1FrontZPosition + fWing1Length/2;
    new G4PVPlacement(0, G4ThreeVector(0, 0, displ), fWing1_log, 
                      "shieldWing1", fContainer_log, false, 0, fCheckOverlaps);


    // ------------------------------- //
    //             WING 2              //
    // ------------------------------- //

    fWing2_sol = new G4Tubs("shieldWing2_sol", 
                             fAluFrameOuterDiameter/2, 
                             fWing2OuterDiameter/2,
                             fWing2Length/2,
                             startAngle=0*deg, 
                             spanningAngle=360*deg);
  
    fWing2_log = new G4LogicalVolume(fWing2_sol, pmma, "shieldWing2_log");
  
    displ = fWing2FrontZPosition + fWing2Length/2;
    new G4PVPlacement(0, G4ThreeVector(0, 0, displ), fWing2_log, 
                      "shieldWing2", fContainer_log, false, 0, fCheckOverlaps);


    // visualization attributes
    VisAttributes();    


    // Set the logical volume
    this -> SetLogicalVolume(fContainer_log);
    
    // Move the detector so that 'tlate' coincides with the
    // center of the foil
    this -> SetTranslation(tlate);
    
    // Check for overlaps
    if (fCheckOverlaps) this -> CheckOverlaps();
}


void VKShield::SetDefaults(){

    fCheckOverlaps = fDetConstr -> GetCheckOverlaps();

    // dimensions
    fAluFrameLength        =  410*mm;
    fAluFrameOuterDiameter =   60*mm;
    fAluFrameThickness     =    3*mm;
    fWindowLength          =   35*mm;
    fEndcapLength          =   10*mm; 
    fHeavyElem1Length      =  200*mm;
    fWing1Length           =   35*mm;
    fWing1OuterDiameter    = fDetConstr -> GetShieldDiameter();
    fWing1FrontZPosition   =  -75*mm;
    fWing2Length           =   35*mm;
    fWing2OuterDiameter    = fDetConstr -> GetShieldDiameter();
    fWing2FrontZPosition   =   40*mm; 

    // deduce container size
    fContainerDiameter = TMath::Max(fAluFrameOuterDiameter, fWing1OuterDiameter);
    fContainerDiameter = TMath::Max(fContainerDiameter, fWing2OuterDiameter);
    fContainerLength   = fAluFrameLength;
}


void VKShield::VisAttributes() {

    // Alu frame
    G4VisAttributes* aluFrame_va = new G4VisAttributes(G4Colour(0.0, 0.0, 1.0)); // blue
    aluFrame_va   -> SetVisibility(true);
    aluFrame_va   -> SetForceWireframe(true);
    fAluFrame_log -> SetVisAttributes(aluFrame_va);

    // Window
    G4VisAttributes* window_va = new G4VisAttributes(G4Colour(1.0, 1.0, 1.0)); // white
    window_va   -> SetVisibility(true);
    window_va   -> SetForceSolid(true);
    fWindow_log -> SetVisAttributes(window_va);
    
    // Endcap
    G4VisAttributes* endcap_va = new G4VisAttributes(G4Colour(1.0, 1.0, 1.0)); // white
    endcap_va   -> SetVisibility(true);
    endcap_va   -> SetForceSolid(true);
    fEndcap_log -> SetVisAttributes(endcap_va);

    // Heavy element 1
    G4VisAttributes* heavyElem1_va = new G4VisAttributes(G4Colour(0.2, 0.0, 0.5)); // ??
    heavyElem1_va   -> SetVisibility(true);
    heavyElem1_va   -> SetForceSolid(true);
    fHeavyElem1_log -> SetVisAttributes(heavyElem1_va);

    // Wing 1
    G4VisAttributes* wing1_va = new G4VisAttributes(G4Colour(1.0, 1.0, 1.0)); // white
    wing1_va   -> SetVisibility(true);
    wing1_va   -> SetForceWireframe(true);
    fWing1_log -> SetVisAttributes(wing1_va);

    // Wing 2
    G4VisAttributes* wing2_va = new G4VisAttributes(G4Colour(1.0, 1.0, 1.0)); // white
    wing2_va   -> SetVisibility(true);
    wing2_va   -> SetForceWireframe(true);
    fWing2_log -> SetVisAttributes(wing2_va);
}
